# Copyright 2018 miruka
# This file is part of pixcat, licensed under LGPLv3.

import io
import math
import random
import re
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, Generator, Optional, Tuple, Union

from PIL import Image as PILImage
from dataclasses import InitVar, dataclass, field

from .terminal import TERM


@dataclass
class Image:
    min_id   = 1
    max_id   = 4_294_967_295
    used_ids = set()
    types    = Union[bytes, str, Path, PILImage.Image]

    source: InitVar[types]
    id:     Optional[int] = None

    origin: types = field(init=False, default=None)

    _pil_image: PILImage.Image = field(init=False, repr=False, default=None)

    _resized_cache: Dict[Tuple[str, tuple], "Image"] = \
        field(init=False, repr=False, compare=False, default_factory=dict)


    def __post_init__(self, source) -> None:
        self._resized_cache = {}  # to make pylint shut up
        self.origin         = source
        self.id             = self._get_id()
        self._pil_image     = self._get_pil_image(source)


    def _get_id(self) -> int:
        # Avoid hanging if somehow more than 4 billion of ids are registered:
        if len(self.used_ids) >= self.max_id:
            self.used_ids = set()

        random_id = random.randint(self.min_id, self.max_id)

        while random_id in self.used_ids:
            random_id = random.randint(self.min_id, self.max_id)

        self.used_ids.add(random_id)
        return random_id


    def _get_pil_image(self, source) -> PILImage.Image:
        if isinstance(source, PILImage.Image):
            return source

        if re.match(r"https?://.+", str(source)):
            import requests
            req = requests.get(source)
            req.raise_for_status()  # Raise if 400 < http code < 600
            source = req.content    # bytes

        if isinstance(source, bytes):
            # Don't use `with`  here, or _get_kitty_file() will fail.
            out = io.BytesIO()
            out.write(source)
            out.seek(0)
            return PILImage.open(out)

        self.origin = path = Path(source).expanduser().resolve()
        return PILImage.open(path)


    def _get_kitty_file(self) -> str:
        dest = NamedTemporaryFile(prefix=".pixcat-", delete=False)
        self._pil_image.save(dest.name, format="PNG", compress_level=0)
        return dest.name


    @property
    def cols(self) -> int:
        return math.ceil(self._pil_image.size[0] / TERM.cell_px_width)

    @property
    def rows(self) -> int:
        return math.ceil(self._pil_image.size[1] / TERM.cell_px_height)


    @staticmethod
    def _negative_col_to_px(num: int) -> int:
        return num if num >= 0 else TERM.cell_px_width * abs(num)

    @staticmethod
    def _negative_row_to_px(num: int) -> int:
        return num if num >= 0 else TERM.cell_px_height * abs(num)


    def resize(self,
               min_w:    int           = 1,
               min_h:    int           = 1,
               max_w:    Optional[int] = None,
               max_h:    Optional[int] = None,
               stretch:  bool          = False,
               resample: str           = "lanczos") -> "Image":

        w, h = img_w, img_h = self._pil_image.size

        max_w = max_w or img_w
        max_h = max_h or img_h

        assert min_w <= max_w
        assert min_h <= max_h

        min_w = self._negative_col_to_px(min_w)
        min_h = self._negative_row_to_px(min_h)
        max_w = self._negative_col_to_px(max_w)
        max_h = self._negative_row_to_px(max_h)

        # Upscale if image is smaller than minimum width/height:
        if (img_w < min_w or img_h < min_h) and img_w < max_w and img_h <max_h:

            if stretch:
                w, h = min_w, min_h

            elif min_w >= min_h:
                # If calculated height > max_h: max_h, if < min_h: min_h
                h = min(max_h, math.ceil((min_w / img_w) * img_h))
                w = math.floor((h / img_h) * img_w)

            else:
                w = min(max_w, math.ceil((min_h / img_h) * img_w))
                h = math.floor((w / img_w) * img_h)

        # Downscale if image is bigger than maximum width/height:
        elif img_w > max_w or img_h > max_h:

            if stretch:
                w, h = max_w, max_h

            elif max_w >= max_h:
                h = min(max_h, math.ceil((max_w / img_w) * img_h))
                w = math.floor((h / img_h) * img_w)

            else:
                w = min(max_w, math.ceil((max_h / img_h) * img_w))
                h = math.floor((w / img_w) * img_h)

        # Nothing to do:
        else:
            return self

        # If an image was already made for decided width/height, return it:

        cached = self._resized_cache.get((w, h))
        if cached:
            return cached

        # Return and save in the cache dict an Image object of the resized.

        resample = getattr(PILImage, resample.upper())
        image    = type(self)(self._pil_image.resize((w, h), resample))

        self._resized_cache[(w, h)] = image
        return image


    def thumbnail(self,
                  size:     int  = 256,
                  stretch:  bool = False,
                  resample: str  = "lanczos") -> "Image":

        return self.resize(*(size,) * 4, stretch, resample)


    def fit_screen(self,
                   h_margin: int  = 0,
                   v_margin: int  = 0,
                   enlarge:  bool = False,
                   stretch:  bool = False,
                   resample: str  = "lanczos") -> "Image":

        h_margin = self._negative_col_to_px(h_margin) * 4
        v_margin = self._negative_row_to_px(v_margin) * 4

        max_wh = (TERM.px_width - h_margin, TERM.px_height - v_margin)
        min_wh =  max_wh if enlarge else (0, 0)

        return self.resize(*min_wh, *max_wh, stretch, resample)


    def show(self,
             x:          Optional[int] = None,
             y:          Optional[int] = None,
             z:          int  = 0,
             relative_x: int  = 0,
             relative_y: int  = 0,
             align:      str  = "center",
             offset_x:   int  = 0,
             offset_y:   int  = 0,
             crop_w:     int  = 0,
             crop_h:     int  = 0) -> "Image":

        # If x/y or relative_(x/y) is set, only one must be set
        assert {bool(x), bool(relative_x)} != {True}
        assert {bool(y), bool(relative_y)} != {True}

        assert align in ("left", "center", "right")

        crop_w = self._negative_col_to_px(crop_w)
        crop_h = self._negative_row_to_px(crop_h)

        params = {
            "offset_x": offset_x, "offset_y": offset_y,
            "crop_w":   crop_w,   "crop_h":   crop_h,
            "z_index":  z,

            "action": "transmit+display",
            "medium" : "tempfile",
            "format" : "png",
            "id":      self.id,
            "payload": self._get_kitty_file(),
        }

        if align == "center":
            relative_x += round(TERM.width / 2) - round(self.cols / 2)

        elif align == "right":
            relative_x += TERM.width - self.cols

        if relative_x:
            print(" " * relative_x, end="")
        elif x:
            print(TERM.move_x(x))  # FIXME

        if relative_y:
            print("\n" * relative_y, end="")  # FIXME: moving up
        elif y:
            print(TERM.move_y(y))  # FIXME

        TERM.run_code(**params)
        return self


    def hide(self, resized_too: bool = True) -> "Image":
        ids = [self.id]

        if resized_too:
            ids += [img.id for img in self._resized_cache.values()]

        for id_ in ids:
            TERM.run_code(action="delete", del_data_target="id", id=id_)

        return self


    def copy(self, new_id: Optional[int] = None) -> "Image":
        return type(self)(source=self.origin, id=new_id)


    def __copy__(self) -> "Image":
        return self.copy()


    @classmethod
    def factory(cls,
                *sources:      "Image.types",
                raise_errors:  bool = False,
                print_errors:  bool = True) -> Generator["Image", None, None]:

        for source in sources:
            try:
                if isinstance(source, (bytes, PILImage.Image)) or \
                   re.match(r"https?://.+", str(source)):
                    yield cls(source)
                    continue

                path = Path(source).expanduser().resolve()

                if path.is_dir():
                    for item in path.iterdir():
                        yield from cls.factory(
                            item,
                            raise_errors = raise_errors,
                            print_errors = print_errors
                        )
                    continue

                yield cls(path)

            except Exception as err:
                if raise_errors:
                    raise

                if print_errors:
                    print(TERM.red("%s: %s" % (type(err).__name__, err)))



import array
import base64
import fcntl
import sys
import termios
from contextlib import contextmanager
from typing import Tuple

import blessed

from . import data


class KittyAnswerError(Exception):
    def __init__(self, from_code: str, answer: str):
        super().__init__(f"{from_code!r} : terminal responded with {answer!r}")


class PixTerminal(blessed.Terminal):
    actions_with_answer = data.ACTIONS_WITH_ANSWER
    img_controls        = data.IMAGE_CONTROLS
    esc                 = data.ESC


    @property
    def size(self) -> Tuple[int, int]:
        return (self.width, self.height)


    @property
    def px_size(self) -> Tuple[int, int]:
        buf = array.array("H", [0, 0, 0, 0])
        fcntl.ioctl(sys.stdout, termios.TIOCGWINSZ, buf)
        return (buf[2], buf[3])

    @property
    def px_width(self) -> int:
        return self.px_size[0]

    @property
    def px_height(self) -> int:
        return self.px_size[1]


    @property
    def cell_px_size(self) -> Tuple[int, int]:
        return (self.px_width // self.width, self.px_height // self.height)

    @property
    def cell_px_width(self) -> int:
        return self.cell_px_size[0]

    @property
    def cell_px_height(self) -> int:
        return self.cell_px_size[1]


    def get_code(self, payload: str = "", **controls: str) -> str:
        real_keys = {
            self.img_controls[k][0]:
                self.img_controls[k][1][v] if self.img_controls[k][1] else v

            for k, v in controls.items()
        }

        keys_str = ",".join([f"{k}={v}" for k, v in real_keys.items()])

        if payload:
            payload = str(base64.b64encode(bytes(payload, "utf-8")), "utf-8")

        # print("%r" % f"{ESC}_G{keys_str};{payload}{ESC}\\")
        return f"{self.esc}_G{keys_str};{payload}{self.esc}\\"


    def run_code(self, payload: str = "", **controls: str) -> None:
        code = self.get_code(payload, **controls)
        print(code)

        if controls.get("action", "transmit") not in self.actions_with_answer:
            return

        # Catch responses kitty print on stdin:
        chars = []
        while True:
            with self.cbreak():
                char = sys.stdin.read(1)
                chars.append(char)
                if char == "\\":
                    break

        self._handle_response(code, "".join(chars))


    @staticmethod
    def _handle_response(from_code: str, answer: str) -> None:
        if answer and ";OK" not in answer:
            raise KittyAnswerError(from_code, answer)


    # y then x for those because blessings does it like that for some reason
    def move_relative(self, y: int = 0, x: int = 0) -> str:
        cursor_y, cursor_x = self.get_location()
        return self.move(cursor_y + y, cursor_x + x)

    def move_relative_x(self, x: int = 0) -> str:
        cursor_y, cursor_x = self.get_location()
        return self.move(cursor_y, cursor_x + x)

    def move_relative_y(self, y: int = 0) -> str:
        cursor_y, cursor_x = self.get_location()
        return self.move(cursor_y + y, cursor_x)


    @contextmanager
    def location_relative(self, x: int = 0, y: int = 0) -> str:
        cursor_y, cursor_x = self.get_location()
        with self.location(x=cursor_x + x, y=cursor_y + y):
            yield


    def align(self, text: str, align: str = "left") -> str:
        if align == "left":
            return self.ljust(text)

        if align == "center":
            return self.center(text)

        if align == "right":
            return self.rjust(text)

        raise ValueError("Alignement must be 'left', 'center' or 'right'.")


TERM = PixTerminal()



# Copyright 2018 miruka
# This file is part of pixcat, licensed under LGPLv3.

"""Usage: pixcat [r|resize | t|thumbnail | f|fit-screen] [options] LOCATION...

Display images on a kitty terminal with optional resizing.

For resizing options, measures (INT number arguments) are in pixel by default.
To indicate columns/rows instead, prefix the number with a dash.

Positioning options are always in columns/rows.

Arguments:
  LOCATION: File, folder to be be scanned recursively for images, or URL.
            Any number of file, folder or URLs can be specified.

Options:
  Resizing:
    -S, --stretch             Do not force keeping the original aspect ratio.
    -r ALGO, --resample ALGO  From fastest/worse to slowest/best quality:
                              nearest, bilinear, bicubic, lanczos (default).

  Specific to r/resize:
    -w INT, --min-width INT   Upscale when width is lower than INT.
    -h INT, --min-height INT  Upscale when height is lower than INT.
    -W INT, --max-width INT   Downscale when height is higher than INT.
    -H INT, --max-height INT  Downscale when height is higher than INT.

  Specific to t/thumbnail:
    -s INT, --size INT  Scale to INTxINT, default 256.

  Specific to f/fit-screen:
    -e, --enlarge                    Scale up for images smaller than terminal.
    -o INT, --horizontal-margin INT  Have a left-right padding of INT columns.
    -v INT, --vertical-margin INT    Have a top-bottom padding of INT columns.

  Positioning:
    -x INT, --absolute-x INT  Left image origin in columns, from terminal left.
    -y INT, --absolute-y INT  Top image origin in rows, from terminal top.
    -z INT, --z-index INT     Images are drawn in front of others that have a
                              lower index. -1 and lower will draw behind text.

    -X INT, --relative-x INT  Like -x, but from current cursor position.
    -Y INT, --relative-y INT  Like -y, but from current cursor position.

    -a ALI, --align ALI       Image and text alignement, -X is added to it.
                              left, center (default) or right.

    -f INT, --offset-x INT    Left offset in pixel, max is column width.
    -F INT, --offset-y INT    Top offset in pixel, max is row height.

    -c INT, --crop-w INT      Crop image left-to-right to INT pixels.
    -C INT, --crop-h INT      Crop image top-to-bottom to INT pixels.

  Extra:
    -O, --print-origin    Print image origin, like a path or URL.
    -n, --print-name      Print image filename.
    -i, --print-id        Print kitty image ID.

    -q, --quiet           Keep quiet about errors, e.g. "cannot identify image"
    -R, --raise-errors    Exit and show full traceback if an error happens.

    -g, --hang            Wait for an enter keypress between every image.
    -G, --hang-final      Wait for enter keypress after all images are drawn.

  Generic:
    --         Mark the end of options, useful if a LOCATION starts by a dash.
    --help     Show this help.
    --version  Show the program version.

Examples:
  pixcat image.jpg
    Just display image.jpg, no resizing.
    If the width exceeds the terminal's, it will be cropped.

  pixcat fit-screen --enlarge image.jpg
    Display image.jpg, downscale or upscale (if --enlarge) when necessary to
    make image match terminal dimensions.

  pixcat resize -w 64 -h 32 -W 512 -H 256 --align right --relative-x -2 *.png
    Display all png files in the current dir with a size of at least 64x32 and
    no bigger than 512x256.

    Align the pictures against the terminal's right edge,
    minus 2 columns (negative --relative-x).

  pixcat thumbnail --size 128 --resample nearest dir1 dir2
    Recursively find and display images in dir1 and dir2,
    with a min-max-size of 128x128 (aspect ratio will be taken into account).

    The nearest rescaling algorithm will trade quality for speed, useful
    if there are a lot of images to display.

  pixcat t -s 128 -r nearest dir1 dir2
    Same as the command above, short form.

Bugs and limitations:
  - Absolute positioning options do not work reliably yet
  - No support in multiplexers like tmux yet
  - Resizing the terminal can lead to a mess, use clear/CTRL+L to fix it."""


import sys
from pathlib import Path
from typing import List, Optional

import docopt

from . import Image, data
from .__about__ import __version__
from .terminal import TERM


def main(argv: Optional[List[str]] = None) -> None:
    argv = argv if argv is not None else sys.argv[1:]

    try:
        params = docopt.docopt(__doc__, argv=argv, version=__version__)
    except docopt.DocoptExit:
        if len(sys.argv) > 1:
            print("Invalid command syntax, check help:\n")

        main(["--help"])
        sys.exit(1)

    images = Image.factory(
        *params["LOCATION"],
        raise_errors = params["--raise-errors"],
        print_errors = not params["--quiet"]
    )

    for image in images:
        handle_image(image, params)

    if params["--hang-final"]:
        input("Press enter to exit...")


def handle_image(image: Image, params: dict) -> None:
    if params["r"] or params["resize"]:
        image = image.resize(**cli_to_func_params("resize", params))

    elif params["t"] or params["thumbnail"]:
        image = image.thumbnail(**cli_to_func_params("thumbnail", params))

    elif params["f"] or params["fit-screen"]:
        image = image.fit_screen(**cli_to_func_params("fit_screen", params))

    print_align = lambda t: print(TERM.align(t, params["--align"] or "center"))

    if params["--print-name"]:
        if isinstance(image.origin, (Path, str)):
            print_align(Path(image.origin).name)
        else:
            print_align("-")

    if params["--print-origin"]:
        if isinstance(image.origin, (Path, str)):
            print_align(str(image.origin))
        else:
            print_align("-")

    if params["--print-id"]:
        print_align(image.id)

    image.show(**cli_to_func_params("show", params))

    if params["--hang"]:
        input()


def cli_to_func_params(func_name: str, params: dict) -> dict:
    mappings = data.CLI_TO_FUNCTIONS_PARAMS
    return {
        mappings[func_name][param][0]: mappings[func_name][param][1](value)
        for param, value in params.items()
        if value is not None and param in mappings[func_name]
    }



# Copyright 2018 miruka
# This file is part of pixcat, licensed under LGPLv3.

"""Display images on a kitty terminal with optional resizing."""

__pkg_name__ = "pixcat"
__version__  = "0.1.3"
__status__   = "Development"
# __status__ = "Production"

__author__   = "miruka"
__email__    = "miruka@disroot.org"

__license__  = "LGPLv3"
